#ifndef CHUNK_H
#define CHUNK_H

#include <cuda_runtime.h>
#include <execinfo.h>

extern cudaStream_t STREAMS[NR_GPUS];
extern int CUR_DEVICE;

extern int CE_SEND_LOCKS[NR_GPUS];
extern int CE_RECV_LOCKS[NR_GPUS];

extern bool ENABLE_CHUNKING;
extern bool USE_DYNAMIC_GROUP_LOCKS;
extern size_t CHUNK_SIZE;

extern cudaStream_t STREAMS[NR_GPUS];

struct ce_lock_state
{	
	int locks[2];
	size_t num_locks;
	size_t budget_remaining;
	bool locked;
	
	ce_lock_state(int device_a, enum cudaMemcpyKind kind, size_t size, int device_b = -1) {
		num_locks = (device_a != -1) + (device_b != -1);
		
		if(device_a != -1) {
			locks[0] = (kind == cudaMemcpyHostToDevice || (kind == cudaMemcpyDeviceToDevice && device_b == -1)) ?
				CE_SEND_LOCKS[device_a] : CE_RECV_LOCKS[device_a];
		}
		
		if(device_b != -1) {
			assert(kind == cudaMemcpyDeviceToDevice);
			
			locks[1] = CE_RECV_LOCKS[device_b];
			
			if(locks[1] < locks[0]) {
				int temp = locks[1];
				locks[1] = locks[0];
				locks[0] = temp;
			}
		}
		else {
			locks[1] = -1;
		}
		
//		if(kind == cudaMemcpyHostToDevice)
//			printf("SENDING: %d\n", size);
//		else if(kind == cudaMemcpyDeviceToHost)
//			printf("RECEVING: %d\n", size);
		
		if(!ENABLE_CHUNKING)
			budget_remaining = size;
		else
			budget_remaining = CHUNK_SIZE;
	}
	
	void crash(void) {
		void *array[50];
		int size, i;
		char **messages;
		
		size = backtrace(array, 50);
		messages = backtrace_symbols(array, size);
		
		fprintf(stderr, "%d: TRIED TO GRAB SAME LOCK TWICE! Lock = %d\n", getpid(), locks[0]);
		for (i = 1; i < size && messages != NULL; ++i)
		{
			fprintf(stderr, "%d: [bt]: (%d) %s\n", getpid(), i, messages[i]);
		}
		free(messages);
		
		assert(false);
	}

	
	void lock() {
		//assert(locks[0] != locks[1]);
		if(locks[0] == locks[1]) crash();
		
		if(USE_DYNAMIC_GROUP_LOCKS) {
			litmus_dgl_lock(locks, num_locks);
		}
		else
		{
			for(int l = 0; l < num_locks; ++l)
			{
				litmus_lock(locks[l]);
			}
		}
		locked = true;
	}
	
	void unlock() {
		if(locks[0] == locks[1]) crash();
		
		if(USE_DYNAMIC_GROUP_LOCKS) {
			litmus_dgl_unlock(locks, num_locks);
		}
		else
		{
			// reverse order
			for(int l = num_locks - 1; l >= 0; --l)
			{
				litmus_unlock(locks[l]);
			}
		}
		locked = false;
	}
	
	void refresh() {
		budget_remaining = CHUNK_SIZE;
	}
	
	bool budgetIsAvailable(size_t tosend) {
		return(tosend >= budget_remaining);
	}
	
	void decreaseBudget(size_t spent) {
		budget_remaining -= spent;
	}
};

// precondition: if do_locking == true, locks in state are held.
static cudaError_t __chunkMemcpy(void* a_dst, const void* a_src, size_t count,
						  enum cudaMemcpyKind kind,
						  ce_lock_state* state)
{
    cudaError_t ret = cudaSuccess;
    int remaining = count;
    
    char* dst = (char*)a_dst;
    const char* src = (const char*)a_src;
    
	// disable chunking, if needed, by setting chunk_size equal to the
	// amount of data to be copied.
	int chunk_size = (ENABLE_CHUNKING) ? CHUNK_SIZE : count;
	int i = 0;
	
    while(remaining != 0)
    {
        int bytesToCopy = std::min(remaining, chunk_size);
		
		if(state && state->budgetIsAvailable(bytesToCopy) && state->locked) {
			cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
			ret = cudaGetLastError();
			
			if(ret != cudaSuccess)
			{
				break;
			}
			
			state->unlock();
			state->refresh(); // replentish.
							  // we can only run out of
							  // budget if chunking is enabled.
							  // we presume that init budget would
							  // be set to cover entire memcpy
							  // if chunking were disabled.			
		}
		
		if(state && !state->locked) {
			state->lock();
		}
		
        //ret = cudaMemcpy(dst+i*chunk_size, src+i*chunk_size, bytesToCopy, kind);
		cudaMemcpyAsync(dst+i*chunk_size, src+i*chunk_size, bytesToCopy, kind, STREAMS[CUR_DEVICE]);
		
		if(state) {
			state->decreaseBudget(bytesToCopy);
		}
		
		//		if(ret != cudaSuccess)
		//		{
		//			break;
		//		}		
		
        ++i;
        remaining -= bytesToCopy;
    }
    return ret;
}

static cudaError_t chunkMemcpy(void* a_dst, const void* a_src, size_t count,
						enum cudaMemcpyKind kind,
						int device_a = -1,  // device_a == -1 disables locking
						bool do_locking = true, 
						int device_b = -1)
{
	cudaError_t ret;
	if(!do_locking || device_a == -1) {
		ret = __chunkMemcpy(a_dst, a_src, count, kind, NULL);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();
	}
	else {
		ce_lock_state state(device_a, kind, count, device_b);
		state.lock();
		ret = __chunkMemcpy(a_dst, a_src, count, kind, &state);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();
		state.unlock();
	}
	return ret;
}


static cudaError_t __chunkMemcpy2D(void* a_dst, size_t dpitch,
							const void* a_src, size_t spitch,
							size_t width, size_t height,
							enum cudaMemcpyKind kind,
							ce_lock_state* state)
{
    cudaError_t ret = cudaSuccess;
    
    char* dst = (char*)a_dst;
    const char* src = (const char*)a_src;
    
    if(dpitch == width && spitch == width)
    {
        ret = __chunkMemcpy(dst, src, width*height, kind, state);
    }
    else
    {
        // need to go row-by-row
        for(size_t i = 0; i < height; ++i)
        {
            size_t dst_offset = i*dpitch;
            size_t src_offset = i*spitch;
            
            char* dst_row_start = dst + dst_offset;
            const char* src_row_start = src + src_offset;
            
            ret = __chunkMemcpy(dst_row_start, src_row_start, width, kind, state);
            
            if(ret != cudaSuccess)
            {
                break;
            }            
        }
    }
    
    return(ret);
}

static cudaError_t chunkMemcpy2D(void* a_dst, size_t dpitch,
						  const void* a_src, size_t spitch,
						  size_t width, size_t height,
						  enum cudaMemcpyKind kind,
						  int device_a = -1,  // device_a == -1 disables locking
						  bool do_locking = true, 
						  int device_b = -1)
{
	cudaError_t ret;
	if(!do_locking || device_a == -1) {
		ret = __chunkMemcpy2D(a_dst, dpitch, a_src, spitch, width, height, kind, NULL);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();		
	}
	else {
		ce_lock_state state(device_a, kind, width*height, device_b);
		state.lock();
		ret = __chunkMemcpy2D(a_dst, dpitch, a_src, spitch, width, height, kind, &state);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();		
		state.unlock();
	}
	return ret;	
}

static cudaError_t __chunkMemcpy3D(const struct cudaMemcpy3DParms *p,
							ce_lock_state* state)
{	
    cudaError_t ret = cudaSuccess;
    
    assert(p->dstArray == NULL && p->srcArray == NULL);
    
    if(p->dstPtr.pitch == p->dstPtr.xsize && p->srcPtr.pitch == p->srcPtr.xsize)
    {
        ret = __chunkMemcpy(p->dstPtr.ptr, p->srcPtr.ptr,
							p->extent.width * p->extent.height * p->extent.depth,
							p->kind, state);
    }
    else
    {
        char* dst = (char*)p->dstPtr.ptr;
        const char* src = (const char*)p->srcPtr.ptr;
        
        size_t dst_plane_size = p->extent.height*p->dstPtr.pitch;
        size_t src_plane_size = p->extent.height*p->srcPtr.pitch;
        
        // need to go plane-by-plane
        for(size_t d = 0; d < p->extent.depth; ++d)
        {
            size_t dst_offset = d*dst_plane_size;
            size_t src_offset = d*src_plane_size;
            
            ret = __chunkMemcpy2D(dst + dst_offset, p->dstPtr.pitch,
								  src + src_offset, p->srcPtr.pitch,
								  p->extent.width, p->extent.height,
								  p->kind, state);
            
            if(ret != cudaSuccess)
            {
                break;
            }
        }
    }
    
    return(ret);
}

static cudaError_t chunkMemcpy3D(const struct cudaMemcpy3DParms *p,
						  int device_a = -1,  // device_a == -1 disables locking
						  bool do_locking = true, 
						  int device_b = -1)
{
	cudaError_t ret;
	if(!do_locking || device_a == -1) {
		ret = __chunkMemcpy3D(p, NULL);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();		
	}
	else {
		ce_lock_state state(device_a, p->kind, p->extent.width * p->extent.height * p->extent.depth, device_b);
		state.lock();
		ret = __chunkMemcpy3D(p, &state);
		cudaStreamSynchronize(STREAMS[CUR_DEVICE]);
		if(ret == cudaSuccess)
			ret = cudaGetLastError();		
		state.unlock();
	}
	return ret;	
}

#endif
